%{
import sure.gfmath.GaloisField;
import tcl.lang.Interp;
import tcl.lang.ReflectObject;
import tcl.lang.TclException;
import tcl.lang.TclObject;
import tcl.lang.TclString;
%}

%token GF_CONSTANT
%token ALG_CONSTANT
%token MATRIX
%token INDESC_X

%token OP_KRON

%token FUNC_TR

%token NUM
%left '-' '+'
%left '*' '/' OP_KRON '%'
%left NEG
%right '^'
%right '('

%%

result
	: expression {
		result = $1;
	}

expression
	: GF_CONSTANT {
		$$ = $1;
	}
	| ALG_CONSTANT {
		$$ = $1;
	}
	| MATRIX {
		$$ = $1;
	}
	| INDESC_X {
		$$ = $1;
	}
	| '(' expression ')' {
		$$ = $2;
	}
	| expression '+' expression {
		$$ = gfCalc.add($1, $3);
	}
	| expression '-' expression {
		$$ = gfCalc.sub($1, $3);
	}
	| expression '*' expression {
		$$ = gfCalc.mul($1, $3);
	}
	| expression '/' expression {
		$$ = gfCalc.div($1, $3);
	}
	| expression '%' expression {
		$$ = gfCalc.mod($1, $3);
	}
	| expression OP_KRON expression {
		$$ = gfCalc.kronecker($1, $3);
	}
	| expression '^' alg_expression {
		$$ = gfCalc.pow($1, $3);
	}	
	| '-' expression %prec NEG {
		$$ = gfCalc.neg($2);
	}
	| '~' expression %prec NEG {
		$$ = gfCalc.inv($2);
	}
	| expression '(' ALG_CONSTANT ')'
	| expression '(' ALG_CONSTANT ',' ALG_CONSTANT ')'
	| FUNC_TR '(' expression ')' {
		$$ = gfCalc.tr($3);
	}

alg_expression
	: ALG_CONSTANT {
		$$ = $1;
	}
	| '(' alg_expression ')' {
		$$ = $2;
	}
	| '-' alg_expression %prec NEG {
		$$ = new GFExprValue(-$2.asInteger());
	}
	
%%

private String err;
private GFExprTokenizer tokenizer;

private GFExprCalculator gfCalc;
private GFExprValue result;

private void yyerror(String s) {
	if(err==null)
		err = s;
	else
		err = err + "; " +s;
}

private int yylex() {
	try {
		GFExprToken t = tokenizer.getNextToken();
		if(t!=null) {
			if(t.getValue()!=null) {
				yylval = t.getValue();
			}
			return t.getId();
		}
	}
	catch(Exception ex) {
		yyerror(ex.getMessage());
	}
	return 0;
}

public TclObject parse(Interp interp, GaloisField gf, String source) throws TclException {
	err = null;
	gfCalc = new GFExprCalculator(gf);
	result = null;
	
	tokenizer = new GFExprTokenizer(interp, gf);
	tokenizer.startString(source);
	try {
		if(yyparse()!=0)
			return null;
		if(result.isMatrix())
			return ReflectObject.newInstance(interp, result.asMatrix().getClass(), result.asMatrix());
		if(result.isPoly())
			return ReflectObject.newInstance(interp, result.asPoly().getClass(), result.asPoly());
		else if(result.isGFElement(gf))
			return TclString.newInstance(gf.elementToString(result.asGFElement()));
		return null;
	}
	finally {
		if(err!=null)
			throw new TclException(interp, "gfexpr parse error ("+err+")");
	}
}


